#!/usr/bin/env python3
#
# Formats C++ code using clang-format.
#

import os
import sys
import subprocess
import argparse

parser = argparse.ArgumentParser(
    description="Format C/C++ code using clang-format."
)
parser.add_argument(
    "operation",
    choices=["check", "format"],
    help='Operation to perform: "check" to check for formatting errors, or "format" to fix formatting errors.',
)
parser.add_argument(
    "--extensions",
    "-e",
    nargs="+",
    default=[".cpp", ".cc", ".c", ".h", ".hpp"],
    help="List of file extensions to check or format (default: .cpp .cc .c .h .hpp)",
)

args = parser.parse_args()


ROOT_DIR = (
    subprocess.check_output(["git", "rev-parse", "--show-toplevel"])
    .strip()
    .decode("utf-8")
)

exclude_dirs = ('.git', '.cache', 'build', 'dependencies', 'docs')
file_list = [os.path.join(dirpath, filename)
             for dirpath, _, filenames in os.walk(ROOT_DIR)
             for filename in filenames
             if any(filename.endswith(ext) for ext in args.extensions)
             and not any(exclude_dir in dirpath for exclude_dir in exclude_dirs)]


def clang_check(file_path: str) -> None:
    try:
        diff_output = subprocess.check_output(
            ["clang-format", "-output-replacements-xml", "-style=file", file_path], stderr=subprocess.STDOUT,
        )
        if b"<replacement " in diff_output:
            print(f"Error: {file_path} needs formatting")
            sys.exit(1)

    except subprocess.CalledProcessError as error:
        print(f'Error: {error.output.decode("utf-8")}')
        sys.exit(1)


def clang_format(file_path: str) -> None:
    diff_output = subprocess.check_output(
        ["clang-format", "-output-replacements-xml", "-style=file", file_path], stderr=subprocess.STDOUT,
    )

    if b"<replacement " in diff_output:
        print(f"Formatting: {file_path}")
        try:
            subprocess.check_call(["clang-format", "-i", "-style=file", file_path])
        except subprocess.CalledProcessError as error:
            print(f'Error: {error.output.decode("utf-8")}')
            sys.exit(1)


def clang_format_verify() -> str:
    version_output = subprocess.check_output(
        ["clang-format", "--version"], stderr=subprocess.STDOUT,
    ).decode("utf-8").split(" ")
    if "version" in version_output :
        return version_output[version_output.index("version") + 1]

    return ""


clang_format_version = clang_format_verify()

for file_path in file_list:
    if args.operation == "check":
        clang_check(file_path)
    elif args.operation == "format":
        clang_format(file_path)


print("Done!")



#!/usr/bin/env python3

import os
from github import Github
import lib.release as release

WORK_DIR = os.path.dirname(os.path.abspath(__file__)).replace("/tools/release", "")

NEXT_TAG = os.environ["NEXT_RELEASE_TAG"]
REPO_NAME = os.environ["GITHUB_REPOSITORY"]
TOKEN = os.environ["GITHUB_TOKEN"]
if not NEXT_TAG or not REPO_NAME or not TOKEN:
    raise Exception(
        "Bad environment variables. Invalid GITHUB_REPOSITORY, GITHUB_TOKEN or NEXT_RELEASE_TAG"
    )

g = Github(TOKEN)
repo = g.get_repo(REPO_NAME)

release_notes = release.contruct_release_notes(repo, NEXT_TAG)

release.create_release(repo, NEXT_TAG, release_notes)

release = repo.get_release(NEXT_TAG)
release.upload_asset("singleheader/ada.cpp")
release.upload_asset("singleheader/ada.h")
release.upload_asset("singleheader/ada_c.h")
release.upload_asset("singleheader/singleheader.zip")



#!/usr/bin/env python3
#
# Creates the amalgamated source files.
#

import zipfile
import sys
import os.path
import subprocess
import os
import re
import shutil
import datetime
if sys.version_info[0] < 3:
    sys.stdout.write("Sorry, requires Python 3.x or better\n")
    sys.exit(1)

SCRIPTPATH = os.path.dirname(os.path.abspath(sys.argv[0]))
PROJECTPATH = os.path.dirname(SCRIPTPATH)
print(f"SCRIPTPATH={SCRIPTPATH} PROJECTPATH={PROJECTPATH}")

if "AMALGAMATE_SOURCE_PATH" not in os.environ:
    AMALGAMATE_SOURCE_PATH = os.path.join(PROJECTPATH, "src")
else:
    AMALGAMATE_SOURCE_PATH = os.environ["AMALGAMATE_SOURCE_PATH"]
if "AMALGAMATE_INCLUDE_PATH" not in os.environ:
    AMALGAMATE_INCLUDE_PATH = os.path.join(PROJECTPATH, "include")
else:
    AMALGAMATE_INCLUDE_PATH = os.environ["AMALGAMATE_INCLUDE_PATH"]
if "AMALGAMATE_OUTPUT_PATH" not in os.environ:
    AMALGAMATE_OUTPUT_PATH = os.path.join(SCRIPTPATH)
else:
    AMALGAMATE_OUTPUT_PATH = os.environ["AMALGAMATE_OUTPUT_PATH"]

# this list excludes the "src/generic headers"
ALLCFILES = ["ada.cpp"]

# order matters
ALLCHEADERS = ["ada.h"]

found_includes = []

current_implementation=''

def doinclude(fid: str, file: str, line: str, origin: str) -> None:

    p = os.path.join(AMALGAMATE_INCLUDE_PATH, file)
    pi = os.path.join(AMALGAMATE_SOURCE_PATH, file)

    if os.path.exists(p):
        if file not in found_includes:
            found_includes.append(file)
            dofile(fid, AMALGAMATE_INCLUDE_PATH, file)
        else:
            pass
    elif os.path.exists(pi):
        if file not in found_includes:
            found_includes.append(file)
            dofile(fid, AMALGAMATE_SOURCE_PATH, file)
        else:
            pass
    else:
        # If we don't recognize it, just emit the #include
        print("unrecognized:", file, " from ", line, " in ", origin)
        print(line, file=fid)

def dofile(fid: str, prepath: str, filename: str) -> None:
    file = os.path.join(prepath, filename)
    RELFILE = os.path.relpath(file, PROJECTPATH)
    # Last lines are always ignored. Files should end by an empty lines.
    print(f"/* begin file {RELFILE} */", file=fid)
    includepattern = re.compile('\\s*#\\s*include "(.*)"')
    with open(file, 'r') as fid2:
        for line in fid2:
            line = line.rstrip('\n')
            s = includepattern.search(line)
            if s:
                includedfile = s.group(1)
                if includedfile == "ada.h" and filename == "ada.cpp":
                    print(line, file=fid)
                    continue

                if includedfile.startswith('../'):
                    includedfile = includedfile[2:]
                # we explicitly include ada headers, one time each
                doinclude(fid, includedfile, line, filename)
            else:
                print(line, file=fid)
    print(f"/* end file {RELFILE} */", file=fid)


# Get the generation date from git, so the output is reproducible.
# The %ci specifier gives the unambiguous ISO 8601 format, and
# does not change with locale and timezone at time of generation.
# Forcing it to be UTC is difficult, because it needs to be portable
# between gnu date and busybox date.
try:
    timestamp = subprocess.run(['git', 'show', '-s', '--format=%ci', 'HEAD'],
                           stdout=subprocess.PIPE).stdout.decode('utf-8').strip()
except:
    print("git not found, timestamp based on current time")
    timestamp = str(datetime.datetime.now())
print(f"timestamp is {timestamp}")

os.makedirs(AMALGAMATE_OUTPUT_PATH, exist_ok=True)
AMAL_H = os.path.join(AMALGAMATE_OUTPUT_PATH, "ada.h")
AMAL_C = os.path.join(AMALGAMATE_OUTPUT_PATH, "ada.cpp")
DEMOCPP = os.path.join(AMALGAMATE_OUTPUT_PATH, "cpp")
README = os.path.join(AMALGAMATE_OUTPUT_PATH, "README.md")

print(f"Creating {AMAL_H}")
amal_h = open(AMAL_H, 'w')
print(f"/* auto-generated on {timestamp}. Do not edit! */", file=amal_h)
for h in ALLCHEADERS:
    doinclude(amal_h, h, f"ERROR {h} not found", h)

amal_h.close()
print()
print()
print(f"Creating {AMAL_C}")
amal_c = open(AMAL_C, 'w')
print(f"/* auto-generated on {timestamp}. Do not edit! */", file=amal_c)
for c in ALLCFILES:
    doinclude(amal_c, c, f"ERROR {c} not found", c)

amal_c.close()

# copy the README and DEMOCPP
if SCRIPTPATH != AMALGAMATE_OUTPUT_PATH:
  shutil.copy2(os.path.join(SCRIPTPATH,"demo.cpp"),AMALGAMATE_OUTPUT_PATH)
  shutil.copy2(os.path.join(SCRIPTPATH,"demo.c"),AMALGAMATE_OUTPUT_PATH)
  shutil.copy2(os.path.join(SCRIPTPATH,"README.md"),AMALGAMATE_OUTPUT_PATH)

shutil.copy2(os.path.join(AMALGAMATE_INCLUDE_PATH,"ada_c.h"),AMALGAMATE_OUTPUT_PATH)

zf = zipfile.ZipFile(os.path.join(AMALGAMATE_OUTPUT_PATH,'singleheader.zip'), 'w', zipfile.ZIP_DEFLATED)
zf.write(os.path.join(AMALGAMATE_OUTPUT_PATH,"ada.cpp"), "ada.cpp")
zf.write(os.path.join(AMALGAMATE_OUTPUT_PATH,"ada.h"), "ada.h")
zf.write(os.path.join(AMALGAMATE_INCLUDE_PATH,"ada_c.h"), "ada_c.h")


print("Done with all files generation.")

print(f"Files have been written to directory: {AMALGAMATE_OUTPUT_PATH}/")
print("Done with all files generation.")




from .. import versions
import os


def test_update_cmakelists_version():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    sample_path = f"{current_dir}/samples/cmakelists.txt"
    sample_expected_path = f"{current_dir}/samples/cmakelists_expected.txt"

    versions.update_cmakelists_version("2.0.0", sample_path)

    with open(sample_path, "r") as cmake:
        given = cmake.read()

    with open(sample_expected_path, "r") as cmake_expected:
        expected = cmake_expected.read()

    assert given == expected
    versions.update_cmakelists_version("1.0.0", sample_path)  # cleanup


def test_update_ada_version_h():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    sample_path = f"{current_dir}/samples/ada_version_h.txt"
    sample_expected_path = f"{current_dir}/samples/ada_version_h_expected.txt"

    versions.update_ada_version_h("2.0.0", sample_path)

    with open(sample_path, "r") as ada_version_h:
        given = ada_version_h.read()

    with open(sample_expected_path, "r") as ada_version_h_expected:
        expected = ada_version_h_expected.read()

    assert given == expected
    versions.update_ada_version_h("1.0.0", sample_path)  # cleanup


def test_update_doxygen_version():
    current_dir = os.path.dirname(os.path.abspath(__file__))
    sample_path = f"{current_dir}/samples/doxygen.txt"
    sample_expected_path = f"{current_dir}/samples/doxygen_expected.txt"

    versions.update_doxygen_version("2.0.0", sample_path)

    with open(sample_path, "r") as doxygen:
        given = doxygen.read()

    with open(sample_expected_path, "r") as doxygen_expected:
        expected = doxygen_expected.read()

    assert given == expected
    versions.update_ada_version_h("1.0.0", sample_path)  # cleanup



#!/usr/bin/env python3

import fileinput
import re


def update_cmakelists_version(new_version: str, file_path: str) -> None:
    inside_project = False
    with fileinput.FileInput(file_path, inplace=True) as cmakelists:
        for line in cmakelists:
            if "set(ADA_LIB_VERSION" in line:
                line = re.sub(r"[0-9]+\.[0-9]+\.[0-9]+", new_version, line)
            elif "set(ADA_LIB_SOVERSION" in line:
                line = re.sub(r"[0-9]+", new_version.split(".")[0], line)

            elif "project(" in line:
                inside_project = True
            elif inside_project:
                if "VERSION" in line:
                    line = re.sub(r"[0-9]+\.[0-9]+\.[0-9]+", new_version, line)
                    inside_project = False
            print(line, end="")


def update_ada_version_h(new_version: str, file_path: str) -> None:
    new_version_list = new_version.split(".")
    with fileinput.FileInput(file_path, inplace=True) as ada_version_h:
        inside_enum = False
        for line in ada_version_h:
            if "#define ADA_VERSION" in line:
                line = f'#define ADA_VERSION "{new_version}"\n'

            elif "enum {" in line:
                inside_enum = True
            elif inside_enum:
                if line.strip().startswith("ADA_VERSION_MAJOR"):
                    line = re.sub(r"\d+", new_version_list[0], line)
                elif line.strip().startswith("ADA_VERSION_MINOR"):
                    line = re.sub(r"\d+", new_version_list[1], line)
                elif line.strip().startswith("ADA_VERSION_REVISION"):
                    line = re.sub(r"\d+", new_version_list[2], line)

            print(line, end="")


def update_doxygen_version(new_version: str, file_path: str) -> None:
    with fileinput.FileInput(file_path, inplace=True) as doxygen:
        for line in doxygen:
            if line.strip().startswith("PROJECT_NUMBER         ="):
                line = f'PROJECT_NUMBER         = "{new_version}"\n'

            print(line, end="")



from .. import release
from datetime import datetime
from collections import namedtuple

Release = namedtuple("Release", ["title", "created_at"])
User = namedtuple("User", ["login"])
Commit = namedtuple("Commit", ["author", "commit"])
CommitMessage = namedtuple("CommitMessage", ["message"])
PullRequestBase = namedtuple("PullRequestBase", "ref")
PullRequestTuple = namedtuple(
    "PullRequest",
    ["title", "number", "state", "base", "merged", "merged_at", "user", "commits"],
)


class PullRequest(PullRequestTuple):
    def get_commits(self):
        return self.commits


class RepoStub:
    def __init__(self):
        self.created_at = datetime(2023, 1, 1)
        self.full_name = "ada-url/ada"

    @staticmethod
    def get_releases() -> list:
        return [
            Release("v1.0.1", datetime(2023, 2, 1)),
            Release("v1.0.3", datetime(2023, 4, 1)),
            Release("v1.0.2", datetime(2023, 3, 1)),
        ]

    @staticmethod
    def get_pulls(state="closed"):
        return list(
            filter(
                lambda pull: pull.state == state,
                [
                    PullRequest(
                        title="Feature 1",
                        number=1,
                        state="open",
                        merged=False,
                        base=PullRequestBase("main"),
                        merged_at=datetime(2023, 2, 2),
                        user=User("contributor_1"),
                        commits=[
                            Commit(
                                User("contributor_1"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("contributor_1"),
                                CommitMessage("src: sample commit 2"),
                            ),
                            Commit(
                                User("contributor_1"),
                                CommitMessage("src: sample commit 3"),
                            ),
                            Commit(
                                User("contributor_1"),
                                CommitMessage("src: sample commit 4"),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="Feature 2",
                        number=2,
                        state="closed",
                        merged=True,
                        merged_at=datetime(2023, 2, 1),
                        base=PullRequestBase("main"),
                        user=User("contributor_2"),
                        commits=[
                            Commit(
                                User("contributor_2"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("contributor_2"),
                                CommitMessage("src: sample commit 2"),
                            ),
                            Commit(
                                User("contributor_2"),
                                CommitMessage("src: sample commit 3"),
                            ),
                            Commit(
                                User("contributor_2"),
                                CommitMessage(
                                    "Co-authored-by: old_contrib_coauthor2 <the@email>"
                                ),
                            ),
                            Commit(
                                User("contributor_2"),
                                CommitMessage("src: sample commit 4"),
                            ),
                            Commit(
                                User("contributor_2"),
                                CommitMessage(
                                    "Co-authored-by: old_contrib_coauthor <the@email>"
                                ),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="Feature 3",
                        number=3,
                        state="closed",
                        merged=True,
                        merged_at=datetime(2023, 2, 2),
                        base=PullRequestBase("main"),
                        user=User("contributor_3"),
                        commits=[
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 2"),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 3"),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 4"),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="Feature 4",
                        number=4,
                        state="closed",
                        merged=True,
                        merged_at=datetime(2023, 2, 3),
                        base=PullRequestBase("main"),
                        user=User("contributor_4"),
                        commits=[
                            Commit(
                                User("contributor_4"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("contributor_4"),
                                CommitMessage("src: sample commit 2"),
                            ),
                            Commit(
                                User("contributor_4"),
                                CommitMessage("src: sample commit 3"),
                            ),
                            Commit(
                                User("contributor_4"),
                                CommitMessage("src: sample commit 4"),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="Feature 5",
                        number=5,
                        state="closed",
                        merged=False,
                        merged_at=datetime(2023, 2, 4),
                        base=PullRequestBase("main"),
                        user=User("contributor_3"),
                        commits=[
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 2"),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 3"),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 4"),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="Feature 6",
                        number=12,
                        state="closed",
                        merged=True,
                        merged_at=datetime(2023, 2, 5),
                        base=PullRequestBase("main"),
                        user=User("contributor_2"),
                        commits=[
                            Commit(
                                User("contributor_2"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("contributor_2"),
                                CommitMessage("src: sample commit 2"),
                            ),
                            Commit(
                                User("contributor_2"),
                                CommitMessage("src: sample commit 3"),
                            ),
                            Commit(
                                User("contributor_2"),
                                CommitMessage("src: sample commit 4"),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="Feature 9",
                        number=13,
                        state="closed",
                        merged=True,
                        merged_at=datetime(2023, 5, 2),
                        base=PullRequestBase("main"),
                        user=User("new_contributor_2"),
                        commits=[
                            Commit(
                                User("new_contributor_2"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("new_contributor_2"),
                                CommitMessage("src: sample commit 2"),
                            ),
                            Commit(
                                User("new_contributor_2"),
                                CommitMessage("src: sample commit 3"),
                            ),
                            Commit(
                                User("new_contributor_2"),
                                CommitMessage("src: sample commit 4 "),
                            ),
                            Commit(
                                User("new_contributor_2"),
                                CommitMessage(
                                    "Co-authored-by: new_contributor_coauthor1 <the@email>"
                                ),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="Feature 7",
                        number=14,
                        state="closed",
                        merged=True,
                        merged_at=datetime(2023, 5, 5),
                        base=PullRequestBase("main"),
                        user=User("contributor_3"),
                        commits=[
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 2"),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 3"),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage("src: sample commit 4 "),
                            ),
                            Commit(
                                User("contributor_3"),
                                CommitMessage(
                                    "Co-authored-by: new_contributor_coauthor2 <the@email>"
                                ),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="Feature 8",
                        number=15,
                        state="closed",
                        merged=True,
                        merged_at=datetime(2023, 5, 1),
                        base=PullRequestBase("main"),
                        user=User("new_contributor_1"),
                        commits=[
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage("src: sample commit 2"),
                            ),
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage("src: sample commit 3"),
                            ),
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage(
                                    "Co-authored-by: new_contributor_coauthor4 <the@email>"
                                ),
                            ),
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage("src: sample commit 4 "),
                            ),
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage(
                                    "Co-authored-by: new_contributor_coauthor3 <the@email>"
                                ),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="Feature 11",
                        number=16,
                        state="closed",
                        merged=True,
                        merged_at=datetime(2023, 5, 10),
                        base=PullRequestBase("another_branch"),
                        user=User("new_contributor_1"),
                        commits=[
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage("src: sample commit 2"),
                            ),
                        ],
                    ),
                    PullRequest(
                        title="chore: release v10.0.0",
                        number=17,
                        state="closed",
                        merged=True,
                        merged_at=datetime(2023, 5, 9),
                        base=PullRequestBase("main"),
                        user=User("new_contributor_1"),
                        commits=[
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage("src: sample commit 1"),
                            ),
                            Commit(
                                User("new_contributor_1"),
                                CommitMessage("src: sample commit 2"),
                            ),
                        ],
                    ),
                ],
            )
        )


def test_get_sorted_merged_pulls():
    pulls = RepoStub.get_pulls(state="closed")
    last_release = None

    sorted_merged_pulls = release.get_sorted_merged_pulls(pulls, last_release)

    # Should return all the merged pull requests since there is no previous release
    assert sorted_merged_pulls == sorted(
        [
            pull
            for pull in pulls
            if pull.merged
            and pull.base.ref == "main"
            and not pull.title.startswith("chore: release")
            and not pull.user.login.startswith("github-actions")
        ],
        key=lambda pull: pull.merged_at,
    )


def test_get_last_release():
    releases = RepoStub.get_releases()

    # Should return the latest release
    last_release = release.get_last_release(releases)
    assert last_release.created_at == datetime(2023, 4, 1)

    # Should return None (in case there are no releases yet)
    last_release = release.get_last_release([])
    assert last_release == None


def test_get_old_contributors():
    last_release = release.get_last_release(RepoStub.get_releases())

    old_contributors = release.get_old_contributors(RepoStub.get_pulls(), last_release)

    # Should return contributors until last release, including co-authors
    assert old_contributors == {
        "contributor_2",
        "contributor_3",
        "contributor_4",
        "old_contrib_coauthor",
        "old_contrib_coauthor2",
    }


def test_get_new_contributors():
    last_release = release.get_last_release(RepoStub.get_releases())
    all_pulls = RepoStub.get_pulls()

    # merged pulls after last release
    merged_pulls = release.get_sorted_merged_pulls(all_pulls, last_release)
    old_contributors = release.get_old_contributors(all_pulls, last_release)

    # Should return a List sorted in alphabetic order with only the new contributors since
    # last release
    new_contributors = release.get_new_contributors(old_contributors, merged_pulls)

    assert new_contributors == [
        "new_contributor_1",
        "new_contributor_2",
        "new_contributor_coauthor1",
        "new_contributor_coauthor2",
        "new_contributor_coauthor3",
        "new_contributor_coauthor4",
    ]


def test_whats_changed_md():
    repo_stub = RepoStub()
    last_release = release.get_last_release(RepoStub.get_releases())
    all_pulls = RepoStub.get_pulls()
    # merged pulls after last release
    merged_pulls = release.get_sorted_merged_pulls(all_pulls, last_release)

    whats_changed = release.whats_changed_md(repo_stub.full_name, merged_pulls)

    assert whats_changed == [
        "* Feature 8 by @new_contributor_1, @new_contributor_coauthor3 and @new_contributor_coauthor4 in https://github.com/ada-url/ada/pull/15",
        "* Feature 9 by @new_contributor_2 and @new_contributor_coauthor1 in https://github.com/ada-url/ada/pull/13",
        "* Feature 7 by @contributor_3 and @new_contributor_coauthor2 in https://github.com/ada-url/ada/pull/14",
    ]


def test_new_contributors_md():
    repo_stub = RepoStub()
    last_release = release.get_last_release(RepoStub.get_releases())
    all_pulls = RepoStub.get_pulls()

    merged_pulls = release.get_sorted_merged_pulls(all_pulls, last_release)
    old_contributors = release.get_old_contributors(all_pulls, last_release)
    new_contributors = release.get_new_contributors(old_contributors, merged_pulls)

    # Should return a markdown containing the new contributors and their first contribution
    new_contributors_md = release.new_contributors_md(
        repo_stub.full_name, merged_pulls, new_contributors
    )

    assert new_contributors_md == [
        "* @new_contributor_2 and @new_contributor_coauthor1 made their first contribution in https://github.com/ada-url/ada/pull/13",
        "* @new_contributor_coauthor2 made their first contribution in https://github.com/ada-url/ada/pull/14",
        "* @new_contributor_1, @new_contributor_coauthor3 and @new_contributor_coauthor4 made their first contribution in https://github.com/ada-url/ada/pull/15",
    ]


def test_full_changelog_md():
    repo_stub = RepoStub()
    last_tag = release.get_last_release(repo_stub.get_releases())

    full_changelog = release.full_changelog_md(
        repo_stub.full_name, last_tag.title, "v3.0.0"
    )
    assert (
        full_changelog
        == "**Full Changelog**: https://github.com/ada-url/ada/compare/v1.0.3...v3.0.0"
    )

    full_changelog = release.full_changelog_md(repo_stub.full_name, None, "v3.0.0")
    assert full_changelog is None


def test_contruct_release_notes():
    repo_stub = RepoStub()

    notes = release.contruct_release_notes(repo_stub, "v3.0.0")
    assert (
        notes
        == "## What's changed\n"
        + "* Feature 8 by @new_contributor_1, @new_contributor_coauthor3 and @new_contributor_coauthor4 in https://github.com/ada-url/ada/pull/15\n"
        + "* Feature 9 by @new_contributor_2 and @new_contributor_coauthor1 in https://github.com/ada-url/ada/pull/13\n"
        + "* Feature 7 by @contributor_3 and @new_contributor_coauthor2 in https://github.com/ada-url/ada/pull/14\n"
        + "\n"
        + "## New Contributors\n"
        + "* @new_contributor_2 and @new_contributor_coauthor1 made their first contribution in https://github.com/ada-url/ada/pull/13\n"
        + "* @new_contributor_coauthor2 made their first contribution in https://github.com/ada-url/ada/pull/14\n"
        + "* @new_contributor_1, @new_contributor_coauthor3 and @new_contributor_coauthor4 made their first contribution in https://github.com/ada-url/ada/pull/15\n"
        + "\n"
        + "**Full Changelog**: https://github.com/ada-url/ada/compare/v1.0.3...v3.0.0"
    )


def test_is_valid_tag():
    assert release.is_valid_tag("v1.0.0") is True
    assert release.is_valid_tag("v1.1.1") is True

    assert release.is_valid_tag("v0") is False
    assert release.is_valid_tag("v1.0.0.0") is False
    assert release.is_valid_tag("1.0.0") is False
    assert release.is_valid_tag("1.0.1") is False


def test_multiple_contributors_mention_md():
    contributors = ["contrib1", "contrib2", "contrib3", "contrib4"]

    md_contributors_mention = release.multiple_contributors_mention_md(contributors)
    assert md_contributors_mention == "@contrib1, @contrib2, @contrib3 and @contrib4"

    contributors = ["contrib1"]
    md_contributors_mention = release.multiple_contributors_mention_md(contributors)
    assert md_contributors_mention == "@contrib1"



#!/usr/bin/env python3

import re
from typing import Optional, List, Set, Union, Type
from github.PullRequest import PullRequest
from github.GitRelease import GitRelease
from github.Repository import Repository


def is_valid_tag(tag: str) -> bool:
    tag_regex = r"^v\d+\.\d+\.\d+$"
    return bool(re.match(tag_regex, tag))


def create_release(
    repository: Repository, tag: str, notes: str
) -> Union[None, Type[Exception]]:
    if not is_valid_tag(tag):
        raise Exception(f"Invalid tag: {tag}")

    try:
        repository.create_git_release(
            tag=tag, name=tag, message=notes, draft=False, prerelease=False
        )

    except Exception as exp:
        raise Exception(
            f"create_release: Error creating release/tag {tag}: {exp!s}"
        ) from exp


def get_sorted_merged_pulls(
    pulls: List[PullRequest], last_release: Optional[GitRelease]
) -> List[PullRequest]:
    # Get merged pulls after last release
    if not last_release:
        return sorted(
            (
                pull
                for pull in pulls
                if pull.merged
                and pull.base.ref == "main"
                and not pull.title.startswith("chore: release")
                and not pull.user.login.startswith("github-actions")
            ),
            key=lambda pull: pull.merged_at,
        )

    return sorted(
        (
            pull
            for pull in pulls
            if pull.merged
            and pull.base.ref == "main"
            and (pull.merged_at > last_release.created_at)
            and not pull.title.startswith("chore: release")
            and not pull.user.login.startswith("github-actions")
        ),
        key=lambda pull: pull.merged_at,
    )


def get_pr_contributors(pull_request: PullRequest) -> List[str]:
    contributors = set()
    for commit in pull_request.get_commits():
        commit_message = commit.commit.message
        if commit_message.startswith("Co-authored-by:"):
            coauthor = commit_message.split("<")[0].split(":")[-1].strip()
            contributors.add(coauthor)
        else:
            author = commit.author
            if author:
                contributors.add(author.login)
    return sorted(list(contributors), key=str.lower)


def get_old_contributors(
    pulls: List[PullRequest], last_release: Optional[GitRelease]
) -> Set[str]:
    contributors = set()
    if last_release:
        merged_pulls = [
            pull
            for pull in pulls
            if pull.merged and pull.merged_at <= last_release.created_at
        ]

        for pull in merged_pulls:
            pr_contributors = get_pr_contributors(pull)
            for contributor in pr_contributors:
                contributors.add(contributor)

    return contributors


def get_new_contributors(
    old_contributors: List[str], merged_pulls: List[PullRequest]
) -> List[str]:
    new_contributors = set()
    for pull in merged_pulls:
        pr_contributors = get_pr_contributors(pull)
        for contributor in pr_contributors:
            if contributor not in old_contributors:
                new_contributors.add(contributor)

    return sorted(list(new_contributors), key=str.lower)


def get_last_release(releases: List[GitRelease]) -> Optional[GitRelease]:
    sorted_releases = sorted(releases, key=lambda r: r.created_at, reverse=True)

    if sorted_releases:
        return sorted_releases[0]

    return None


def multiple_contributors_mention_md(contributors: List[str]) -> str:
    contrib_by = ""
    if len(contributors) <= 1:
        for contrib in contributors:
            contrib_by += f"@{contrib}"
    else:
        for contrib in contributors:
            contrib_by += f"@{contrib}, "

        contrib_by = contrib_by[:-2]
        last_comma = contrib_by.rfind(", ")
        contrib_by = (
            contrib_by[:last_comma].strip()
            + " and "
            + contrib_by[last_comma + 1 :].strip()
        )
    return contrib_by


def whats_changed_md(repo_full_name: str, merged_pulls: List[PullRequest]) -> List[str]:
    whats_changed = []
    for pull in merged_pulls:
        contributors = get_pr_contributors(pull)
        contrib_by = multiple_contributors_mention_md(contributors)

        whats_changed.append(
            f"* {pull.title} by {contrib_by} in https://github.com/{repo_full_name}/pull/{pull.number}"
        )

    return whats_changed


def get_first_contribution(
    merged_pulls: List[str], contributor: str
) -> Optional[PullRequest]:
    for pull in merged_pulls:
        contrubutors = get_pr_contributors(pull)
        if contributor in contrubutors:
            return pull

    # ? unreachable
    return None


def new_contributors_md(
    repo_full_name: str, merged_pulls: List[PullRequest], new_contributors: List[str]
) -> List[str]:
    contributors_by_pr = {}
    contributors_md = []
    for contributor in new_contributors:
        first_contrib = get_first_contribution(merged_pulls, contributor)

        if not first_contrib:
            continue

        if first_contrib.number not in contributors_by_pr.keys():
            contributors_by_pr[first_contrib.number] = [contributor]
        else:
            contributors_by_pr[first_contrib.number] += [contributor]

    contributors_by_pr = dict(sorted(contributors_by_pr.items()))
    for pr_number, contributors in contributors_by_pr.items():
        contributors.sort(key=str.lower)
        contrib_by = multiple_contributors_mention_md(contributors)

        contributors_md.append(
            f"* {contrib_by} made their first contribution in https://github.com/{repo_full_name}/pull/{pr_number}"
        )

    return contributors_md


def full_changelog_md(
    repository_name: str, last_tag_name: str, next_tag_name: str
) -> Optional[str]:
    if not last_tag_name:
        return None
    return f"**Full Changelog**: https://github.com/{repository_name}/compare/{last_tag_name}...{next_tag_name}"


def contruct_release_notes(repository: Repository, next_tag_name: str) -> str:
    repo_name = repository.full_name
    last_release = get_last_release(repository.get_releases())
    all_pulls = repository.get_pulls(state="closed")

    sorted_merged_pulls = get_sorted_merged_pulls(all_pulls, last_release)
    old_contributors = get_old_contributors(all_pulls, last_release)
    new_contributors = get_new_contributors(old_contributors, sorted_merged_pulls)

    whats_changed = whats_changed_md(repo_name, sorted_merged_pulls)

    new_contrib_md = new_contributors_md(
        repo_name, sorted_merged_pulls, new_contributors
    )

    notes = "## What's changed\n"
    for changes in whats_changed:
        notes += changes + "\n"

    notes += "\n"

    if new_contributors:
        notes += "## New Contributors\n"
        for new_contributor in new_contrib_md:
            notes += new_contributor + "\n"

        notes += "\n"

    if last_release:
        notes += full_changelog_md(
            repository.full_name, last_release.title, next_tag_name
        )

    return notes



#!/usr/bin/env python3

import os
import lib.versions as update_versions
from lib.release import is_valid_tag

WORK_DIR = os.path.dirname(os.path.abspath(__file__)).replace("/tools/release", "")

ADA_VERSION_H = f"{WORK_DIR}/include/ada/ada_version.h"
DOXYGEN = f"{WORK_DIR}/doxygen"
CMAKE_LISTS = f"{WORK_DIR}/CMakeLists.txt"

NEXT_TAG = os.environ["NEXT_RELEASE_TAG"]
if not NEXT_TAG or not is_valid_tag(NEXT_TAG):
    raise Exception(f"Bad environment variables. Invalid NEXT_RELEASE_TAG {NEXT_TAG}.")

NEXT_TAG = NEXT_TAG[1:]  # from v1.0.0 to 1.0.0

update_versions.update_ada_version_h(NEXT_TAG, ADA_VERSION_H)
update_versions.update_doxygen_version(NEXT_TAG, DOXYGEN)
update_versions.update_cmakelists_version(NEXT_TAG, CMAKE_LISTS)
